------------------------------------------------------------
-- Buffy by Sonaza
------------------------------------------------------------

local ADDON_NAME, SHARED = ...;

local _G = getfenv(0);

local LibStub = LibStub;
local A = LibStub("AceAddon-3.0"):NewAddon(ADDON_NAME, "AceEvent-3.0");
_G[ADDON_NAME] = A;
SHARED[1] = A;

-- Text Strings
BUFFY_SET_BINDING_TEXT 		= "|cffeed028Set Buffy Cast Keybind|r";
BUFFY_CHOOSE_BINDING_TEXT	= "Choose a Binding";
BUFFY_PRESS_BUTTON_TEXT		= "Press a key to use it as the binding";
BUFFY_TEMPORARY_BIND_TEXT	= "Buffy will set a temporary binding when required and you can use any key you want without overwriting existing ones."
BUFFY_CANCEL_BINDING_TEXT 	= "Press Escape to Cancel";
BUFFY_ACCEPT_TEXT			= "Save";

-- Some buff category data
local BUFF_STATS 		= 0x001;
local BUFF_STAMINA 		= 0x002;
local BUFF_ATTACK_POWER	= 0x004;
local BUFF_HASTE		= 0x008;
local BUFF_SPELL_POWER	= 0x010;
local BUFF_CRIT			= 0x020;
local BUFF_MASTERY		= 0x040;
local BUFF_MULTISTRIKE	= 0x080;
local BUFF_VERSATILITY	= 0x100;

local BUFF_SPECIAL		= 0xFFF;

local BUFF_TYPES = {
	BUFF_STATS,
	BUFF_STAMINA,
	BUFF_ATTACK_POWER,
	BUFF_HASTE,
	BUFF_SPELL_POWER,
	BUFF_CRIT,
	BUFF_MASTERY,
	BUFF_MULTISTRIKE,
	BUFF_VERSATILITY,
};
A.BUFF_TYPES = BUFF_TYPES;

local BUFF_TYPE_NAMES = {
	[BUFF_STATS] 		= RAID_BUFF_1,
	[BUFF_STAMINA] 		= RAID_BUFF_2,
	[BUFF_ATTACK_POWER] = RAID_BUFF_3,
	[BUFF_HASTE] 		= RAID_BUFF_4,
	[BUFF_SPELL_POWER] 	= RAID_BUFF_5,
	[BUFF_CRIT] 		= RAID_BUFF_6,
	[BUFF_MASTERY] 		= RAID_BUFF_7,
	[BUFF_MULTISTRIKE] 	= RAID_BUFF_8,
	[BUFF_VERSATILITY] 	= RAID_BUFF_9,	
};
A.BUFF_TYPE_NAMES = BUFF_TYPE_NAMES;

local BUFF_SPELL_CATEGORIES = {
	[BUFF_STATS] 		= {},
	[BUFF_STAMINA] 		= {},
	[BUFF_ATTACK_POWER] = {},
	[BUFF_HASTE] 		= {},
	[BUFF_SPELL_POWER] 	= {},
	[BUFF_CRIT] 		= {},
	[BUFF_MASTERY] 		= {},
	[BUFF_MULTISTRIKE] 	= {},
	[BUFF_VERSATILITY] 	= {},
};
A.BUFF_SPELL_CATEGORIES = BUFF_SPELL_CATEGORIES;

local BUFFS = {};
local BUFF_CATEGORY_BY_SPELL = {};
---------------------------

local function AddBuffSpell(spell, categories, name)
	if(not spell or not categories) then return end
	
	if(name) then
		BUFFS[name] = spell;
	end
	
	BUFF_CATEGORY_BY_SPELL[spell] = categories;
	
	if(categories == BUFF_SPECIAL) then return end
	
	for _, category in pairs(BUFF_TYPES) do
		if(bit.band(categories, category) > 0) then
			tinsert(BUFF_SPELL_CATEGORIES[category], spell);
		end
	end
end

local function GetBuffCategory(spell)
	if(not spell) then return end
	
	return BUFF_CATEGORY_BY_SPELL[spell];
end

AddBuffSpell(20217,		BUFF_STATS, "BLESSING_OF_KINGS");
AddBuffSpell(115921,	BUFF_STATS, "LEGACY_OF_THE_EMPEROR");
AddBuffSpell(1126,		BUFF_STATS + BUFF_VERSATILITY, "MARK_OF_THE_WILD");
AddBuffSpell(116781,	BUFF_STATS + BUFF_CRIT, "LEGACY_OF_THE_WHITE_TIGER");
AddBuffSpell(160206,	BUFF_STATS, "LONEWOLF_PRIMATES");
AddBuffSpell(159988,	BUFF_STATS, "PET_DOG");
AddBuffSpell(160017,	BUFF_STATS, "PET_GORILLA");
AddBuffSpell(90363,		BUFF_STATS + BUFF_CRIT, "PET_SHALE_SPIDER");
AddBuffSpell(160077,	BUFF_STATS + BUFF_VERSATILITY, "PET_WORM");

AddBuffSpell(469, 		BUFF_STAMINA, "COMMANDING_SHOUT");
AddBuffSpell(21562,		BUFF_STAMINA, "POWER_WORD_FORTITUDE");
AddBuffSpell(166928,	BUFF_STAMINA, "BLOOD_PACT");
AddBuffSpell(160199,	BUFF_STAMINA, "LONEWOLF_BEAR");
AddBuffSpell(50256,		BUFF_STAMINA, "PET_BEAR");
AddBuffSpell(160014,	BUFF_STAMINA, "PET_GOAT");
AddBuffSpell(160003,	BUFF_STAMINA + BUFF_HASTE, "PET_RYLAK");
AddBuffSpell(90364,		BUFF_STAMINA + BUFF_SPELL_POWER, "PET_SILITHID");

AddBuffSpell(6673, 		BUFF_ATTACK_POWER, "BATTLE_SHOUT");
AddBuffSpell(57330, 	BUFF_ATTACK_POWER, "HORN_OF_WINTER");
AddBuffSpell(19506, 	BUFF_ATTACK_POWER, "TRUESHOT_AURA");

AddBuffSpell(55610,		BUFF_HASTE + BUFF_VERSATILITY, "UNHOLY_AURA");
AddBuffSpell(49868,		BUFF_HASTE + BUFF_MULTISTRIKE, "MIND_QUICKENING");
AddBuffSpell(116956,	BUFF_HASTE + BUFF_MASTERY, "GRACE_OF_AIR");
AddBuffSpell(113742,	BUFF_HASTE + BUFF_MULTISTRIKE, "SWIFTBLADES_CUNNING");
AddBuffSpell(160203,	BUFF_HASTE, "LONEWOLF_HYENA");
AddBuffSpell(128432,	BUFF_HASTE, "PET_HYENA");
AddBuffSpell(135678,	BUFF_HASTE, "PET_SPOREBAT");
AddBuffSpell(160074,	BUFF_HASTE, "PET_WASP");

AddBuffSpell(1459,		BUFF_SPELL_POWER + BUFF_CRIT, "ARCANE_BRILLIANCE");
AddBuffSpell(61316,		BUFF_SPELL_POWER + BUFF_CRIT, "DALARAN_BRILLIANCE");
AddBuffSpell(109773,	BUFF_SPELL_POWER + BUFF_MULTISTRIKE, "DARK_INTENT");
AddBuffSpell(160205,	BUFF_SPELL_POWER + BUFF_CRIT, "LONEWOLF_SERPENT");
AddBuffSpell(126309,	BUFF_SPELL_POWER, "PET_WATER_STRIDER");
AddBuffSpell(128433,	BUFF_SPELL_POWER, "PET_SERPENT");

AddBuffSpell(17007,		BUFF_CRIT, "LEADER_OF_THE_PACK");
AddBuffSpell(160205,	BUFF_CRIT, "LONEWOLF_RAPTOR");
AddBuffSpell(160052,	BUFF_CRIT, "PET_RAPTOR");
AddBuffSpell(24604,		BUFF_CRIT, "PET_WOLF");
AddBuffSpell(90309,		BUFF_CRIT, "PET_DEVILSAUR");
AddBuffSpell(126373,	BUFF_CRIT, "PET_QUILEN");

AddBuffSpell(155522, 	BUFF_MASTERY, "POWER_OF_THE_GRAVE");
AddBuffSpell(19740, 	BUFF_MASTERY, "BLESSING_OF_MIGHT");
AddBuffSpell(24907, 	BUFF_MASTERY, "MOONKIN_AURA");
AddBuffSpell(160198, 	BUFF_MASTERY, "LONEWOLF_CAT");
AddBuffSpell(93435, 	BUFF_MASTERY, "PET_CAT");
AddBuffSpell(160039, 	BUFF_MASTERY, "PET_HYDRA");
AddBuffSpell(160073, 	BUFF_MASTERY, "PET_TALLSTRIDER");
AddBuffSpell(128997, 	BUFF_MASTERY, "PET_SPIRIT_BEAST");

AddBuffSpell(166916,	BUFF_MULTISTRIKE, "WINDFLURRY");
AddBuffSpell(172968, 	BUFF_MULTISTRIKE, "LONEWOLF_DRAGONHAWK");
AddBuffSpell(50519, 	BUFF_MULTISTRIKE, "PET_BAT");
AddBuffSpell(34889, 	BUFF_MULTISTRIKE, "PET_DRAGONHAWK");
AddBuffSpell(24844, 	BUFF_MULTISTRIKE, "PET_WIND_SERPENT");
AddBuffSpell(58604, 	BUFF_MULTISTRIKE, "PET_CORE_HOUND");
AddBuffSpell(57386, 	BUFF_MULTISTRIKE + BUFF_VERSATILITY, "PET_CLEFTHOOF");

AddBuffSpell(167188,	BUFF_VERSATILITY, "INSPIRING_PRESENCE");
AddBuffSpell(167187,	BUFF_VERSATILITY, "SANCTITY_AURA");
AddBuffSpell(172967,	BUFF_VERSATILITY, "LONEWOLF_RAVAGER");
AddBuffSpell(50518,		BUFF_VERSATILITY, "PET_RAVAGER");
AddBuffSpell(159735,	BUFF_VERSATILITY, "PET_BIRD_OF_PREY");
AddBuffSpell(35290,		BUFF_VERSATILITY, "PET_BOAR");
AddBuffSpell(160045,	BUFF_VERSATILITY, "PET_PORCUPINE");
AddBuffSpell(173035,	BUFF_VERSATILITY, "PET_STAG");

AddBuffSpell(324, 		BUFF_SPECIAL, "SHAMAN_LIGHTNING_SHIELD");
-- AddBuffSpell(52127, 	BUFF_SPECIAL, "SHAMAN_WATER_SHIELD");

AddBuffSpell(2823, 		BUFF_SPECIAL, "ROGUE_DEADLY_POISON");
AddBuffSpell(8679, 		BUFF_SPECIAL, "ROGUE_WOUND_POISON");
AddBuffSpell(3408, 		BUFF_SPECIAL, "ROGUE_CRIPPLING_POISON");
AddBuffSpell(108211,	BUFF_SPECIAL, "ROGUE_LEECHING_POISON");

AddBuffSpell(15473,		BUFF_SPECIAL, "PRIEST_SHADOWFORM");

AddBuffSpell(25780,		BUFF_SPECIAL, "PALADIN_RIGHTEOUS_FURY");

-- RAID_BUFF_1 Stats
-- RAID_BUFF_2 Stamina
-- RAID_BUFF_3 Attack Power
-- RAID_BUFF_4 Haste
-- RAID_BUFF_5 Spell Power
-- RAID_BUFF_6 Critical Strike
-- RAID_BUFF_7 Mastery
-- RAID_BUFF_8 Multistrike
-- RAID_BUFF_9 Versatility

local function CanPoisonWeapons()
	-- return IsEquippedItemType("One-Hand");
	
	local mainhand = GetInventoryItemLink("player", 16);
	local mhSlot = mainhand and select(9, GetItemInfo(mainhand)) or "";
	
	local offhand = GetInventoryItemLink("player", 17);
	local ohSlot = offhand and select(9, GetItemInfo(offhand)) or "";
	
	return mhSlot == "INVTYPE_WEAPON" or mhSlot == "INVTYPE_WEAPONMAINHAND" or
		   ohSlot == "INVTYPE_WEAPON" or ohSlot == "INVTYPE_WEAPONOFFHAND";
end

local CLASS_CASTABLE_BUFFS = {
	["WARRIOR"]	= {
		[0]	= {
			{
				raidbuff = { BUFFS.BATTLE_SHOUT, BUFFS.COMMANDING_SHOUT, },
			},
		},
	},
	["DEATHKNIGHT"]	= {
		[0]	= {
			{
				raidbuff = { BUFFS.HORN_OF_WINTER },
			},
		},
	},
	["PALADIN"]	= {
		[0]	= {
			{ 
				raidbuff = { BUFFS.BLESSING_OF_KINGS, BUFFS.BLESSING_OF_MIGHT },
			},
		},
		[2]	= {
			{ 
				raidbuff = { BUFFS.BLESSING_OF_KINGS, BUFFS.BLESSING_OF_MIGHT },
			},
			{
				selfbuff = { BUFFS.PALADIN_RIGHTEOUS_FURY, },
				condition = function()
					return A.db.global.Class.Paladin.RemindRighteousFury;
				end,
				description = "Missing Righteous Fury",
			},
		},
	},
	["MONK"] = {
		[1]	= {
			{
				raidbuff = { BUFFS.LEGACY_OF_THE_WHITE_TIGER, },
			},
		},
		[2]	= {
			{
				raidbuff = { BUFFS.LEGACY_OF_THE_EMPEROR, },
			},
		},
		[3]	= {
			{
				raidbuff = { BUFFS.LEGACY_OF_THE_WHITE_TIGER, },
			},
		},
	},
	["PRIEST"] = {
		[0]	= {
			{
				raidbuff = { BUFFS.POWER_WORD_FORTITUDE, },
			},
		},
		[3]	= {
			{
				raidbuff = { BUFFS.POWER_WORD_FORTITUDE, },
			},
			{
				selfbuff = function()
					return GetShapeshiftForm() ~= 1, BUFFS.PRIEST_SHADOWFORM;
				end,
				condition = function()
					return A.db.global.Class.Priest.RemindShadowform;
				end,
				description = "Not in Shadowform",
			},
		},
	},
	["SHAMAN"] = {
		[0]	= {
			{
				selfbuff = { BUFFS.SHAMAN_LIGHTNING_SHIELD, },
				condition = function()
					return A.db.global.Class.Shaman.RemindShield;
				end,
			},
		},
	},
	["DRUID"] = {
		[0]	= {
			{
				raidbuff = { BUFFS.MARK_OF_THE_WILD, },
			},
		},
	},
	["ROGUE"] = {
		[0] = {
			{
				selfbuff = { BUFFS.ROGUE_DEADLY_POISON, BUFFS.ROGUE_WOUND_POISON, },
				condition = function()
					return CanPoisonWeapons() and A.db.global.Class.Rogue.EnableLethal and not A.db.global.Class.Rogue.WoundPoisonPriority;
				end,
				description = "Missing Lethal Poison",
			},
			{
				selfbuff = { BUFFS.ROGUE_WOUND_POISON, BUFFS.ROGUE_DEADLY_POISON, },
				condition = function()
					return CanPoisonWeapons() and A.db.global.Class.Rogue.EnableLethal and A.db.global.Class.Rogue.WoundPoisonPriority;
				end,
				description = "Missing Lethal Poison",
			},
			{
				noTalent = {3, 2},
				selfbuff = { BUFFS.ROGUE_CRIPPLING_POISON, },
				condition = function()
					return CanPoisonWeapons() and A.db.global.Class.Rogue.EnableNonlethal and not A.db.global.Class.Rogue.SkipCrippling;
				end,
				description = "Missing Non-Lethal Poison",
			},
			{
				hasTalent = {3, 2},
				selfbuff = { BUFFS.ROGUE_LEECHING_POISON, },
				condition = function()
					return CanPoisonWeapons() and A.db.global.Class.Rogue.EnableNonlethal;
				end,
				description = "Missing Non-Lethal Poison",
			},
		},
	},
	["MAGE"] = {
		[0]	= {
			{
				raidbuff = { BUFFS.DALARAN_BRILLIANCE, BUFFS.ARCANE_BRILLIANCE },
			},
		},
	},
	["WARLOCK"] = {
		[0]	= {
			{
				raidbuff = { BUFFS.DARK_INTENT },
			},
		},
	},
	["HUNTER"] = {
		[0]	= {
			{
				hasTalent = {7, 3},
				raidbuff = {
					BUFFS.LONEWOLF_PRIMATES,
					BUFFS.LONEWOLF_DRAGONHAWK,
					BUFFS.LONEWOLF_RAPTOR,
					BUFFS.LONEWOLF_CAT,
					BUFFS.LONEWOLF_HYENA,
					BUFFS.LONEWOLF_RAVAGER,
					BUFFS.LONEWOLF_BEAR,
					BUFFS.LONEWOLF_SERPENT,
				},
				condition = function()
					return A.db.global.Class.Hunter.EnableLoneWolf;
				end,
			},
		},
		[1]	= {
			-- Beast Mastery doesn't have Lone Wolf
		},
	},
};

function A:OnInitialize()
	SLASH_BUFFY1	= "/bf";
	SLASH_BUFFY2	= "/buffy";
	SlashCmdList["BUFFY"] = function(command)
		A:OpenContextMenu(UIParent);
	end
	
	A:InitializeDatabase();
end

function A:OnEnable()
	A:RegisterEvent("UNIT_AURA");
	A:RegisterEvent("PLAYER_ALIVE");
	A:RegisterEvent("PLAYER_UNGHOST", "PLAYER_ALIVE");
	A:RegisterEvent("PLAYER_TALENT_UPDATE");
	A:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED");
	A:RegisterEvent("PLAYER_EQUIPMENT_CHANGED");
	A:RegisterEvent("PLAYER_REGEN_DISABLED");
	A:RegisterEvent("PLAYER_REGEN_ENABLED");
	A:RegisterEvent("ZONE_CHANGED");
	A:RegisterEvent("ZONE_CHANGED_NEW_AREA", "ZONE_CHANGED");
	A:RegisterEvent("COMPANION_UPDATE");
	
	A:InitializeDatabroker();

	A.IsFrameLocked = true;
	
	A:ScanTalents();
	A:UpdateBuffs();
	
	CreateFrame("Frame"):SetScript("OnUpdate", function(self, elapsed)
		self.elapsed = self.elapsed or 0 + elapsed;
		
		if(self.elapsed >= 5.0) then
			A:UpdateBuffs();
		end
	end);
end

function A:GetRealSpellID(spell_id)
	local spell_name = GetSpellInfo(spell_id);
	local name, _, _, _, _, _, realSpellID = GetSpellInfo(spell_name);
	
	return realSpellID or spell_id;
end

function A:ZONE_CHANGED()
	A:UpdateBuffs();
end

function A:UNIT_AURA()
	A:UpdateBuffs();
end

function A:COMPANION_UPDATE(event, companiontype)
	if(companiontype == "MOUNT") then
		A:UpdateBuffs();
	end
end

function A:PLAYER_SPECIALIZATION_CHANGED()
	A:PLAYER_TALENT_UPDATE();
end

function A:PLAYER_EQUIPMENT_CHANGED()
	A:UpdateBuffs();
end

function A:PLAYER_REGEN_DISABLED()
	A:UpdateBuffs();
	A.IsFrameLocked = true;
	BuffyFrame:EnableMouse(false);
	
	BuffySpellButtonFrame:SetAttribute("type1", nil);
end

function A:PLAYER_REGEN_ENABLED()
	A:UpdateBuffs();
	BuffyFrame:EnableMouse(true);
end

function A:PLAYER_TALENT_UPDATE()
	A:ScanTalents();
	A:UpdateBuffs();
end

function A:PLAYER_ALIVE()
	A:UpdateBuffs();
end

function A:ScanTalents()
	A.ActiveTalents = {};
	
	local numTiers = GetMaxTalentTier();
	local group = GetActiveSpecGroup();
	for tier = 1, numTiers do
		for column = 1, 3 do
			local talentID, _, _, selected = GetTalentInfo(tier, column, group);
			if(selected) then
				A.ActiveTalents[tier] = column;
				break;
			end
		end
	end	
end

function A:PlayerHasTalent(tier, column)
	return A.ActiveTalents and A.ActiveTalents[tier] and A.ActiveTalents[tier] == column;
end

function A:UnitHasBuff(unit, spell)
	if(not unit or not spell) then return false end
	
	local realSpellID = A:GetRealSpellID(spell);
	local spell_name = GetSpellInfo(realSpellID);
	if(not spell_name) then return false end
	
	local name, _, _, _, debuffType, duration, expirationTime, unitCaster = UnitAura(unit, spell_name, nil, "HELPFUL");
	if(not name) then
		return false;
	end
	
	return true, unitCaster == "player", unitCaster, expirationTime - GetTime();
end

function A:UnitInRange(unit)
	if(not unit) then return false end
	
	if(unit == "player") then return not UnitIsDeadOrGhost(unit) end
	
	if(not UnitIsVisible(unit) or not UnitIsConnected(unit) or UnitIsDeadOrGhost(unit)) then
		return false;
	end
	
	local unitInRange = nil;
	local distance, distanceChecked = UnitDistanceSquared(unit);
	if(distanceChecked) then
		unitInRange = distance <= 900;
	end
	
	return unitInRange;
end

local GROUP_TYPE_SOLO 	= 1;
local GROUP_TYPE_PARTY 	= 2;
local GROUP_TYPE_RAID	= 3;

function A:GetGroupType()
	if(IsInRaid()) then
		return GROUP_TYPE_RAID;
	elseif(IsInGroup()) then
		return GROUP_TYPE_PARTY;
	end
	
	return GROUP_TYPE_SOLO;
end

local partyUnitID = { "player", "party1", "party2", "party3", "party4" };

function A:GetUnitID(group_type, index)
	if(group_type == GROUP_TYPE_SOLO or group_type == GROUP_TYPE_PARTY) then
		return partyUnitID[index];
	elseif(group_type == GROUP_TYPE_RAID) then
		return string.format("raid%d", index);
	end
	
	return nil;
end

local function GroupIterator()
	local index = 0;
	local groupType = A:GetGroupType();
	local numGroupMembers = GetNumGroupMembers();
	if(groupType == GROUP_TYPE_SOLO) then numGroupMembers = 1 end
	
	-- print(index, group_type, numGroupMembers);
	
	return function()
		index = index + 1;
		if(index <= numGroupMembers) then
			return index, A:GetUnitID(groupType, index);
		end
	end
end

function A:UnitIsPet(searchunit)
	if(not searchunit) then return false end
	
	local pets = {};
	for index, unit in GroupIterator() do
		if(UnitExists(unit .. "pet") and UnitIsUnit(searchunit, unit .. "pet")) then
			return true, UnitName(unit), UnitName(unit .. "pet");
		end
	end
	
	return false;
end

function A:ScanBuffs()
	local missingBuffs = {};
	
	for category, spells in pairs(BUFF_SPELL_CATEGORIES) do
		for index, unit in GroupIterator() do
			local buffFound = false;
			local isCastByPlayer;
			
			for _, spell in ipairs(spells) do
				local hasBuff, isInRange;
				hasBuff, isCastByPlayer = A:UnitHasBuff(unit, spell);
				
				if(hasBuff) then
					buffFound = true;
					break;
				end
			end
			
			if(not buffFound) then
				if(not missingBuffs[unit]) then
					missingBuffs[unit] = {};
				end
				
				tinsert(missingBuffs[unit], category);
			end
		end
	end
	
	return missingBuffs;
end

function A:ScanCurrentBuffs()
	local buffs = {};
	
	for category, spells in pairs(BUFF_SPELL_CATEGORIES) do
		local hasBuff, isCastByPlayer, buff_spell, caster = A:UnitHasCategoryBuff("player", category);
		
		if(not hasBuff) then
			buffs[category] = false;
		else
			local name, class;
			if(caster ~= nil) then
				name = UnitName(caster);
				_, class = UnitClass(caster);
			else
				name = "Unknown";
				class = "PRIEST";
			end
			
			buffs[category] = {
				buff = buff_spell,
				casterName = name,
				casterClass = class,
			};
		end
	end
	
	return buffs;
end

function A:ScanMissingBuff(spell)
	if(not spell) then return nil end
	
	local spellCategory = BUFF_CATEGORY_BY_SPELL[spell];
	if(not spellCategory) then return nil end
	
	local checkCategories = {};
	
	for _, category in pairs(BUFF_TYPES) do
		if(bit.band(spellCategory, category) > 0) then
			tinsert(checkCategories, category);
		end
	end
	
	local unitsMissingBuffs = {};
	local buffCastByPlayer;
	
	for _, category in ipairs(checkCategories) do
		for index, unit in GroupIterator() do
			local buffFound = false;
			local isCastByPlayer;
			
			for _, spell in ipairs(BUFF_SPELL_CATEGORIES[category]) do
				local hasBuff;
				hasBuff, isCastByPlayer = A:UnitHasBuff(unit, spell);
				
				if(hasBuff) then
					buffFound = true;
					buffCastByPlayer = isCastByPlayer;
					break;
				end
			end
			
			if(not buffFound and A:UnitInRange(unit)) then
				tinsert(unitsMissingBuffs, unit);
			end
		end
	end
	
	return unitsMissingBuffs, buffCastByPlayer;
end

function A:UnitHasCategoryBuff(unit, category)
	if(not unit or not category) then return nil end
	if(not A:UnitInRange(unit)) then return nil end
	
	local hasBuff, isCastByPlayer, buff_spell;
	
	for _, spell in ipairs(BUFF_SPELL_CATEGORIES[category]) do
		hasBuff, isCastByPlayer, caster = A:UnitHasBuff(unit, spell);
		if(hasBuff) then
			buff_spell = spell;
			break
		end
	end
	
	return hasBuff, isCastByPlayer, buff_spell, caster;
end

local BUFF_STATE_NO_BUFF 				= 0;
local BUFF_STATE_HAS_BUFF 				= 1;
local BUFF_STATE_HAS_BUFF_BY_PLAYER 	= 2;

function A:ScanBuffList(bufflist)
	if(not bufflist or type(bufflist) ~= "table") then return nil end
	
	local partyBuffState = {};
	
	for _, spell in ipairs(bufflist) do
		if(not spell) then return nil end
		
		local spellCategory = BUFF_CATEGORY_BY_SPELL[spell];
		if(not spellCategory) then return nil end
		
		local checkCategories = {};
		
		for _, category in pairs(BUFF_TYPES) do
			if(bit.band(spellCategory, category) > 0) then
				tinsert(checkCategories, category);
			end
		end
		
		for _, category in ipairs(checkCategories) do
			if(not partyBuffState[category]) then
				partyBuffState[category] = {};
			end
			
			for index, unit in GroupIterator() do
				local hasBuff, isCastByPlayer = A:UnitHasCategoryBuff(unit, category);
				
				if(hasBuff ~= nil) then
					local buffState = BUFF_STATE_NO_BUFF;
					if(hasBuff) then
						if(not isCastByPlayer) then
							buffState = BUFF_STATE_HAS_BUFF;
						else
							buffState = BUFF_STATE_HAS_BUFF_BY_PLAYER;
						end
					end
					
					partyBuffState[category][unit] = buffState;
				end
			end
		end
	end
	
	return partyBuffState;
end

-- Buffy:GetPlayerRaidBuffByCategory(1)
function A:GetPlayerRaidBuffByCategory(search_category)
	local _, PLAYER_CLASS = UnitClass("player");
	local spec = GetSpecialization();
	
	-- Try to find a spec based list and if not found default to basic list
	local buffs = CLASS_CASTABLE_BUFFS[PLAYER_CLASS][spec] or CLASS_CASTABLE_BUFFS[PLAYER_CLASS][0];
	if(not buffs) then return false end
	
	for _, data in ipairs(buffs) do
		if(data.raidbuff) then
			for _, spell in ipairs(data.raidbuff) do
				local spellCategory = BUFF_CATEGORY_BY_SPELL[spell];
				if(not spellCategory) then return nil end
				
				if(bit.band(spellCategory, search_category) > 0 and IsSpellKnown(spell)) then
					return spell;
				end
			end
		end
	end
	
	return nil;
end 

function A:UnitHasSomeBuff(unit, bufflist)
	for _, buff in ipairs(bufflist) do
		if(IsSpellKnown(buff)) then
			local hasBuff = A:UnitHasBuff(unit, buff);
			if(hasBuff) then
				return true, buff;
			end
		end
	end
	
	return false, bufflist[1];
end

local function IsReallyMounted()
	-- IsMounted() or 
	return IsFlying() or UnitOnTaxi("player") or UnitInVehicle("player");
end

function A:UpdateBuffs()
	A:UpdateDatabrokerText();
	
	local inInstance, instanceType = A:PlayerInInstance();
	
	local hideAlert = not self.db.global.Enabled or
					  (self.db.global.DisableOutside and not inInstance) or
					  (not self.db.global.ShowInCombat and InCombatLockdown()) or
					  (not self.db.global.ShowWhileMounted and IsReallyMounted());
	
	if(hideAlert) then
		A:HideBuffyAlert();
		return;
	end
	
	local _, PLAYER_CLASS = UnitClass("player");
	local spec = GetSpecialization();
	
	-- Try to find a spec based list and if not found default to basic list
	local buffs = CLASS_CASTABLE_BUFFS[PLAYER_CLASS][spec] or CLASS_CASTABLE_BUFFS[PLAYER_CLASS][0];
	if(not buffs) then return false end
	
	local playerCastExists = false;
	local shouldCastSpell = false;
	local castSpellID;
	
	local infoUnits = {};
	
	for _, data in ipairs(buffs) do
		local valid = true;
		
		if(data.hasTalent) then
			valid = valid and A:PlayerHasTalent(data.hasTalent[1], data.hasTalent[2]);
		end
		
		if(data.noTalent) then
			valid = valid and not A:PlayerHasTalent(data.noTalent[1], data.noTalent[2]);
		end
		
		if(data.condition and type(data.condition) == "function") then
			valid = valid and data.condition();
		end
		
		if(valid) then
			if(data.raidbuff) then
				local buffStates = A:ScanBuffList(data.raidbuff);
				
				local playerBuffCategory = 0;
				local playerBuffExists = false;
				
				for category, unitBuffs in pairs(buffStates) do
					for unit, buffState in pairs(unitBuffs) do
						if(buffState == BUFF_STATE_HAS_BUFF_BY_PLAYER) then
							playerBuffExists = true;
							playerBuffCategory = category;
							
							break;
						end
					end
					
					if(playerBuffExists) then break end
				end
				
				local missingCategory = 0;
				local everybodyHasBuff = true;
				
				if(playerBuffExists) then
					for unit, buffState in pairs(buffStates[playerBuffCategory]) do
						if(buffState ~= BUFF_STATE_HAS_BUFF_BY_PLAYER) then
							everybodyHasBuff = false;
							missingCategory = playerBuffCategory;
							infoUnits[unit] = unit;
						end
					end
					
					if(not everybodyHasBuff) then
						for category, unitBuffs in pairs(buffStates) do
							if(category ~= missingCategory) then
								for unit, buffState in pairs(unitBuffs) do
									if(buffState == BUFF_STATE_HAS_BUFF_BY_PLAYER) then
										infoUnits[unit] = nil;
									end
								end
							end
						end
						
						local unitsMissing = 0;
						for _, _ in pairs(infoUnits) do
							unitsMissing = unitsMissing + 1;
						end
						
						if(unitsMissing == 0) then
							everybodyHasBuff = true;
						end
					end
				else
					for category, unitBuffs in pairs(buffStates) do
						for unit, buffState in pairs(unitBuffs) do
							if(buffState == BUFF_STATE_NO_BUFF) then
								everybodyHasBuff = false;
								missingCategory = category;
								infoUnits[unit] = unit;
							end
						end
						
						if(not everybodyHasBuff) then break end
					end
				end
				
				if(not everybodyHasBuff) then
					local buffSpell = A:GetPlayerRaidBuffByCategory(missingCategory);
					
					local units = {};
					for unit, _ in pairs(infoUnits) do
						tinsert(units, unit);
					end
					
					A:ShowBuffyAlert(buffSpell, {
						units = units,
						category = missingCategory,
					});
					
					return;
				end
			end
			
			if(data.selfbuff and not UnitIsDeadOrGhost("player")) then
				if(type(data.selfbuff) == "function") then
					local shouldAlert, buff = data.selfbuff();
					
					if(shouldAlert and IsSpellKnown(buff)) then
						A:ShowBuffyAlert(buff, {
							description = data.description,
						});
						return;
					end
				else
					local hasBuff, buff = A:UnitHasSomeBuff("player", data.selfbuff);
					
					if(not hasBuff and IsSpellKnown(buff)) then
						A:ShowBuffyAlert(buff, {
							description = data.description,
						});
						return;
					end
				end
			end
		end
	end
	
	A:HideBuffyAlert();
end

function A:ShowBuffyAlert(spell, data)
	if(not spell) then return end
	local realSpellID = A:GetRealSpellID(spell);
	
	local spellName, rank, icon = GetSpellInfo(realSpellID);
	BuffyFrame.icon.texture:SetTexture(icon);
	
	if(not InCombatLockdown()) then
		BuffySpellButtonFrame:SetAttribute("type1", "spell");
		BuffySpellButtonFrame:SetAttribute("spell1", spellName);
		BuffySpellButtonFrame:SetAttribute("target1", "player");
		A:SetTempBind();
	end
	
	BuffyFrame.buffname:SetFormattedText("|cfff1da54Cast|r|n%s", spellName);
	
	BuffyFrame.description:Hide();
	
	if(data) then
		if(data.units and A:GetGroupType() ~= GROUP_TYPE_SOLO) then
			local numUnits = #data.units;
			BuffyFrame.description:SetFormattedText("%d player%s missing %s", numUnits, numUnits == 1 and "" or "s", BUFF_TYPE_NAMES[data.category]);
			BuffyFrame.description:Show();
		elseif(data.description) then
			BuffyFrame.description:SetText(data.description);
			BuffyFrame.description:Show();
		end
	end
	
	if(not BuffyFrame:IsShown()) then
		BuffyFrame.fadein:Play();
	end
end

function A:HideBuffyAlert()
	if(not InCombatLockdown()) then
		BuffySpellButtonFrame:SetAttribute("type1", nil);
		A:ClearTempBind();
	end
	
	if(BuffyFrame:IsShown()) then
		BuffyFrame.fadeout:Play();
	end
end

function A:PlayerInInstance()
	local name, instanceType = GetInstanceInfo();
	
	if(instanceType == "none" or C_Garrison.IsOnGarrisonMap()) then
		return false;
	end
	
	return true, instanceType;
end

function BuffyFrame_OnShow(self)
	local scale = UIParent:GetEffectiveScale();
	self:SetScale(1.0 / scale * 0.92);
	
	BuffyFrame_RestorePosition();
	
	-- self:ClearAllPoints();
	-- self:SetPoint("LEFT", UIParent, "CENTER", -22, GetScreenHeight() * 0.18);
end


function BuffyFrame_OnMouseDown()
	if(A.IsFrameLocked) then return end
	
	CloseMenus();
	
	BuffyFrame:StartMoving();
	BuffyFrame:SetUserPlaced(false);
	BuffyFrame.IsMoving = true;
end

function BuffyFrame_OnMouseUp()
	if(BuffyFrame.IsMoving) then
		BuffyFrame:StopMovingOrSizing();
		BuffyFrame.IsMoving = false;
		BuffyFrame_SavePosition();
	end
end

function BuffyFrame_SavePosition()
	local point, _, relativePoint, x, y = BuffyFrame:GetPoint();
	
	A.db.global.Position = {
		x = x,
		y = y,
		Point = point,
		RelativePoint = relativePoint,
		DefaultPosition = false,
	};
end

function BuffyFrame_RestorePosition()
	BuffyFrame:ClearAllPoints();
	
	if(not A.db.global or A.db.global.Position.DefaultPosition) then
		BuffyFrame:SetPoint("LEFT", UIParent, "CENTER", -22, GetScreenHeight() * 0.18);
	else
		local p = A.db.global.Position;
		BuffyFrame:SetPoint(p.Point, UIParent, p.RelativePoint, p.x, p.y);
	end
end

function BuffyFrame_ResetPosition()
	BuffyFrame:ClearAllPoints();
	BuffyFrame:SetPoint("LEFT", UIParent, "CENTER", -22, GetScreenHeight() * 0.18);
	
	local point, _, relativePoint, x, y = BuffyFrame:GetPoint();
	
	A.db.global.Position = {
		x = x,
		y = y,
		Point = point,
		RelativePoint = relativePoint,
		DefaultPosition = true,
	};
	
	BuffyFrame_RestorePosition();
end

function A:ResetPosition()
	BuffyFrame_ResetPosition();
end

local MESSAGE_PATTERN = "|cff3ebfeaBuffy|r %s";
function A:AddMessage(pattern, ...)
	DEFAULT_CHAT_FRAME:AddMessage(MESSAGE_PATTERN:format(string.format(pattern, ...)));
end

function A:GetBinding()
	-- return GetBindingKey("CLICK BuffySpellButtonFrame:LeftButton");
	return A.db.global.Keybind;
end

function A:SetTempBind()
	local key = A:GetBinding();
	if(key) then
		SetOverrideBindingClick(BuffyFrame, true, key, "BuffySpellButtonFrame", "LeftButton");
	end
end

function A:ClearTempBind()
	ClearOverrideBindings(BuffyFrame);
end

function BuffyKeybindingFrame_OnShow()
	BuffyKeybindingFrame.bindkey = A:GetBinding();
	
	if(not BuffyKeybindingFrame.bindkey) then
		BuffyKeybindingFrame.currentBinding:SetText(BUFFY_CHOOSE_BINDING_TEXT);
	else
		BuffyKeybindingFrame.currentBinding:SetText(BuffyKeybindingFrame.bindkey);
	end
	
	-- BuffyKeybindingFrame.bindingUsedBy:SetText(BUFFY_PRESS_BUTTON_TEXT);
end

function BuffyKeybindingFrame_OnKeydown(self, key)
	if(key == "ESCAPE") then
		BuffyKeybindingFrameOuter:Hide();
	end
	
	if(key == "LSHIFT" or key == "RSHIFT" or key == "LCTRL" or key == "RCTRL" or key == "LALT" or key == "RALT") then
		return;
	end
	
	if(IsShiftKeyDown()) then key = "SHIFT-" .. key; end
	if(IsControlKeyDown()) then key = "CTRL-" .. key; end
	if(IsAltKeyDown()) then key = "ALT-" .. key; end
	
	BuffyKeybindingFrame.bindkey = key;
	BuffyKeybindingFrame.currentBinding:SetText(key);
	
	-- local bindAction = GetBindingAction(key);
	-- local bindActionText = _G["BINDING_NAME_" .. bindAction];
	
	-- if(bindAction ~= "") then
	-- 	BuffyKeybindingFrame.bindingUsedBy:SetFormattedText("%s|n%s", bindAction, bindActionText or "");
	-- else
	-- 	BuffyKeybindingFrame.bindingUsedBy:SetText("Binding not in use");
	-- end
end

function BuffyKeybindingFrame_OnMouseWheel(self, delta)
	if(delta >= 1) then
		BuffyKeybindingFrame_OnKeydown(self, "MOUSEWHEELUP");
	elseif(delta <= -1) then
		BuffyKeybindingFrame_OnKeydown(self, "MOUSEWHEELDOWN");
	end
end

function BuffyKeybindingFrame_OnMouseDown(self, button)
	if(button == "LeftButton") then
		BuffyKeybindingFrame_OnKeydown(self, "BUTTON1");
	elseif(button == "RightButton") then
		BuffyKeybindingFrame_OnKeydown(self, "BUTTON2");
	elseif(button == "MiddleButton") then
		BuffyKeybindingFrame_OnKeydown(self, "BUTTON3");
	else
		BuffyKeybindingFrame_OnKeydown(self, button);
	end
end

function BuffyKeybindingFrameAcceptButton_OnClick()
	if(A:GetBinding()) then A:ClearTempBind(); end
	
	A.db.global.Keybind = BuffyKeybindingFrame.bindkey;
	A:SetTempBind();
	
	BuffyKeybindingFrame:Hide();
end

function BuffySpellButtonFrame_OnClick()
	if(InCombatLockdown()) then
		A:AddMessage("Cannot cast recommended buff while in combat");
	end
end

function BuffyFrame_OnClick(self, button)
	if(button == "RightButton") then
		A:OpenContextMenu(self);
	end
end

function BuffyFrame_OnFadeInPlay(self)
	BuffyFrame:Show();
end

function BuffyFrame_OnFadeOutFinished(self)
	BuffyFrame:Hide();
end